# -*- coding: utf-8 -*-

'''USAQG

SDFI jasodfij aosdif a
'''

import pika
import os
import json
import itertools


class ZZNBus(object):

    def __init__(self, url, rpcExchange='RPC', eventExchange='EVT'):
        self.rpcExchange = rpcExchange
        self.eventExchange = eventExchange
        self.conn = pika.BlockingConnection(pika.URLParameters(url))
        self.channel = self.conn.channel()
        self.channel.add_on_return_callback(self.on_return)
        self.channel.add_on_cancel_callback(self.on_cancel)
        self.channel.add_on_close_callback(self.on_close)
        self._rpc = None

    def on_cancel(self, *args, **kwds):
        print 'CANCEL', args, kwds

    def on_return(self, *args, **kwds):
        print 'RETURN', args, kwds

    def on_close(self, *args, **kwds):
        print 'CLOSE', args, kwds

    def send(self, routing_key, event):
        self.channel.basic_publish(
            exchange=self.eventExchange,
            routing_key=routing_key,
            body=json.dumps(event),
            mandatory=False,
        )

    def monitor(self, bindings, callback):
        queue = self.channel.queue_declare(exclusive=True).method.queue
        for exchange, routing_keys in bindings.items():
            for routing_key in routing_keys:
                print 'exchange=%s routing_key=%s' % (exchange, routing_key)
                self.channel.queue_bind(
                    queue=queue,
                    exchange=exchange,
                    routing_key=routing_key,
                    # nowait=True,
                    # arguments=None
                )
        self.channel.basic_consume(
            consumer_callback=callback,
            queue=queue,
            no_ack=True,
            exclusive=True,
            # consumer_tag=None,
            # arguments=None
        )
        print 'start receiving messages...'
        self.channel.start_consuming()

    def close(self):
        self.channel.close()
        self.conn.close()

    @property
    def rpc(self):
        if self._rpc is None:
            self._rpc = RPCClient(self)
        return self._rpc


class RPCClient(object):

    def __init__(self, bus):
        self._bus = bus
        queue = self._bus.channel.queue_declare(exclusive=True)
        self._queue = queue.method.queue
        self._ctag = self._bus.channel.basic_consume(
            self._consumer, queue=self._queue, no_ack=True
        )
        self._results = {}

    def _consumer(self, ch, method, props, body):
        # print 'CB', method, props, body
        self._results[props.correlation_id] = json.loads(body)

    _idCounter = itertools.count()

    def nextCorrelationId(self):
        return 'req-%d-%s-%d' % (
            os.getpid(), id(self), self._idCounter.next()
        )

    def call(self, requests, callback=None):
        results = {}
        reqmap = {}
        for request in requests:
            request['id'] = cid = self.nextCorrelationId()
            reqmap[cid] = request
            self._bus.channel.basic_publish(
                exchange=self._bus.rpcExchange,
                routing_key=request['method'],
                properties=pika.BasicProperties(
                    reply_to=self._queue,
                    correlation_id=cid,
                ),
                body=json.dumps(request),
                mandatory=True,
            )
        while True:
            for cid, res in self._results.items():
                if callback is not None:
                    callback(reqmap[cid], res)
                results[cid] = res
            self._results.clear()
            if len(results) == len(requests):
                break
            self._bus.conn.process_data_events()
        return [(reqmap[cid], results[cid])
                for cid in set(reqmap) | set(results)]

    def __call__(self, request):
        res = self.call([request])
        return res[0][1]

    def __getattr__(self, name):
        return RpcNamespace(self, name)


class RpcNamespace(object):

    def __init__(self, rpc, name):
        self.__rpc = rpc
        self.__name = name

    def __getattr__(self, method):
        return RpcNamespace(self.__rpc, '%s.%s' % (self.__name, method))

    def __call__(self, *args, **kwds):
        if args:
            assert not kwds, repr((args, kwds))
            params = list(args)
        elif kwds:
            params = [kwds]
        else:
            params = []
        request = dict(
            method=self.__name,
            params=params
        )
        res = self.__rpc(request)
        if res.get('error'):
            raise Exception(res['error'])
        else:
            return res.get('result')


DEFAULT_URL = "amqp://agent:{MQ_PASSWD}@mq.is.uh.in/dat"


def main():
    import logging
    from argparse import ArgumentParser
    # from agent5.loggers import script_logging

    parser = ArgumentParser()
    parser.add_argument(
        '-v', '--verbose', default=0, action='count')
    parser.add_argument(
        '-U', '--url', metavar='URL', default=DEFAULT_URL,
        help='RabbitMQ URL (default %(default)s)')
    parser.add_argument(
        'action', metavar='ACTION',
        help='Action')
    parser.add_argument(
        'args', nargs='*', metavar='ARG',
        help='Action arguments')

    args = parser.parse_args()
    args.url = args.url.format(**os.environ)
    # script_logging(verbosity=args.verbose)
    # logging.getLogger('pika').setLevel(logging.WARNING)

    if args.action == 'shell':
        if args.args:
            parser.error('unwanted args for shell action')
        return action_shell(args.url)
    elif args.action == 'monitor':
        if not args.args:
            bindings = {'EVT': ['#']}
        else:
            bindings = {}
            for arg in args.args:
                bindings.update(json.loads(arg))
        return action_monitor(args.url, bindings)
    else:
        parser.error('unknown action: %s' % args.action)
    return 0


def action_monitor(url, bindings):

    def dump_message(ch, method, props, body):
        print '------------------------------------------------------------'
        print '%s -> %s %s' % (
            method.exchange,
            method.routing_key,
            json.dumps(json.loads(body), indent=2, sort_keys=True)
        )
    bus = ZZNBus(url)
    bus.monitor(bindings, dump_message)


def action_shell(url, histfile='~/.rpc-shell-history'):
    import code
    import sys
    import pprint
    import rlcompleter
    import readline
    import atexit
    sys.displayhook = pprint.pprint
    bus = ZZNBus(url)
    namespace = dict(bus=bus)
    readline.set_completer(rlcompleter.Completer(namespace).complete)
    readline.parse_and_bind("tab: complete")
    if histfile:
        histfile = os.path.expanduser(histfile)
        if os.path.isfile(histfile):
            readline.read_history_file(histfile)

        def savehist():
            readline.set_history_length(100)
            readline.write_history_file(histfile)

        atexit.register(savehist)
    con = code.InteractiveConsole(namespace)
    con.resetbuffer()
    con.interact(banner='''
bus = ZZNBus instance (connected)
''')
